import os
import click
import sys


@click.command()
@click.argument('pipeline_name', default=None, required=False)
def main(pipeline_name):
    """
    Run a pipeline or all pipelines if pipeline_name is None. This is a command line tool to run a pipeline or all pipelines in the current directory.

    Args:
        pipeline_name: Name of the pipeline to run or None
    """
    main_file = "main.py"
    current_directory = os.getcwd()
    main_path = os.path.join(current_directory, main_file)

    if os.path.isfile(main_path):
        if pipeline_name is not None:
            click.secho(f"Directory successfully located. Running pipeline(s) '{pipeline_name}'...", fg='green')
            os.system(f'{sys.executable} -c "from main import run_pipelines; run_pipelines(\'{pipeline_name}\')"')
        else:
            click.secho("Directory successfully located. Running Pipelines...", fg='green')
            os.system(f'{sys.executable} -c "from main import run_pipelines; run_pipelines()"')
    else:
        click.secho(f"{main_file} not found in the current directory.", fg='red')


if __name__ == "__main__":
    main()



from jetline.data.data import Data
from jetline.data.helper import choose_file

"""
Here you can add your Data classes. You can import some helper classes @jetline.data.helper .
Every Dataclass needs a name and a data attributes. The name is like an id so only use a name once.

"""
class YourDataClass(Data):
    """
    Class representing YourDataClass.

    Inherits from Data class.

    Attributes:
        name (str): The name of the data.
        data (str): The data obtained from the file.

    Methods:
        __init__(): Initializes a new instance of the YourDataClass.

    """
    def __init__(self):
        super().__init__(
            name="Name",
            data=choose_file(),

        )



import importlib.util
from pathlib import Path
import toml
import inspect
from jetline.data.data import DataManager
from jetline.logging import logger
import os
import sys


class Pipeline:
    def __init__(self, nodes=None, data_manager=None):
        """
        Initialize the Pipeline with a list of nodes and a data manager.
        """
        self.nodes = {node.name: node for node in nodes} if nodes else {}
        self.data_manager = data_manager
        self.outputs = {}

    def run(self):
        """
        Runs the execute method of each node in the self.nodes dictionary
        and stores the output in the self.outputs dictionary.
        """
        for node_name, node in self.nodes.items():
            self.outputs[node_name] = node.execute(self.data_manager)

    def add_node(self, node):
        """
        Add a single node to the pipeline.
        """
        if node.name in self.nodes:
            raise ValueError(f"Node with name '{node.name}' already exists.")
        self.nodes[node.name] = node


class PipelineManager:
    def __init__(self):
        """
        Initialize the pipeline manager.
        """
        self.caller_path = self.get_caller_path()
        self.add_project_directory_to_sys_path()
        self.data_manager = DataManager(caller_path=self.caller_path)
        self.pipelines = {}
        self.load_pipelines()

    @staticmethod
    def get_caller_path():
        """
        Description:
        This method returns the path of the file that called it.

        Parameters:
        - self: The instance of the calling class. This is an implicit parameter and does not need to be provided explicitly.

        Returns:
        - module_path: The path of the file that called the method.


        Example Usage:
        instance = MyClass()
        path = instance.get_caller_path()

        """
        frame = inspect.stack()[2]
        module = inspect.getmodule(frame[0])
        if module is not None:
            module_path = Path(module.__file__).parent
        else:
            module_path = Path.cwd()
        return module_path

    def add_project_directory_to_sys_path(self):
        """

        Add the project directory to the system path.

        This method is used to add the project directory to the sys.path list in order to make modules and packages from the project directory importable.

        """
        toml_file_path = os.path.join(self.caller_path, 'project.toml')

        with open(toml_file_path, 'r') as f:
            toml_data = toml.load(f)
            place = toml_data.get('project', {}).get('place')

        if not place:
            raise FileNotFoundError('The project.toml file was not found in the directory. Hint: run cd to your project')

        target_directory = os.path.join(self.caller_path, place)

        sys.path.append(target_directory)

    def load_pipelines(self):
        """
        Load pipelines from project.toml file and add them to the pipelines dictionary.

        Args:
            self: The instance of the class calling the method.

        Returns:
            None

        Raises:
            ValueError: If the register function in the pipeline module does not return a Pipeline instance.

        """
        base_path = self.caller_path

        project_toml_path = base_path / "project.toml"
        with open(project_toml_path, "r") as f:
            project_config = toml.load(f)
            pipeline_folder_name = project_config["project"]["place"]
            pipeline_folder = base_path / pipeline_folder_name

            for pipeline_name in pipeline_folder.iterdir():
                if pipeline_name.is_dir():
                    pipeline_name = pipeline_name.name
                    pipeline_module_path = pipeline_folder / pipeline_name / "pipeline.py"
                    if pipeline_module_path.exists():
                        spec = importlib.util.spec_from_file_location("pipeline_module", pipeline_module_path)
                        pipeline_module = importlib.util.module_from_spec(spec)
                        spec.loader.exec_module(pipeline_module)
                        pipeline_register_function = getattr(pipeline_module, "register", None)
                        if callable(pipeline_register_function):

                            pipeline_definition = pipeline_register_function(self.data_manager)
                            if isinstance(pipeline_definition, Pipeline):
                                self.pipelines[pipeline_name] = pipeline_definition
                            else:
                                raise ValueError(
                                    f"The register function in {pipeline_module_path} must return a Pipeline instance.")

    def run(self, pipeline_order):
        """
        Run method to execute pipelines in a specific order.

        Parameters:
        - self: The object instance.
        - pipeline_order (list): The order in which the pipelines should be executed.

        Returns:
        - results (dict): A dictionary with the results of each pipeline execution.

        Raises:
        - ValueError: If any of the pipelines in the specified order are not found.

        Example usage:
            pipeline_order = ['pipeline1', 'pipeline2', 'pipeline3']
            results = obj.run(pipeline_order)
        """
        results = {}

        missing_pipelines = [pipeline_name for pipeline_name in pipeline_order if pipeline_name not in self.pipelines]
        if missing_pipelines:
            error_message = f"Error running pipelines: The following pipelines were not found: {', '.join(missing_pipelines)}"
            logger.error(error_message)
            raise ValueError(error_message)

        for pipeline_name in pipeline_order:
            pipeline = self.pipelines[pipeline_name]
            results[pipeline_name] = pipeline.run()
            logger.info(f"Pipeline '{pipeline_name}' executed successfully.")
        logger.info("All Pipelines ran successfully.")
        return results



import tkinter as tk
from tkinter import filedialog
import pandas as pd
import os


def choose_file():
    """
    This method opens a file dialog window to allow the user to choose a file on their system. It returns the path of the selected file.
    """
    root = tk.Tk()
    root.withdraw()
    file_path = filedialog.askopenfilename()
    return file_path


def load_file_to_dataframe(file_path, delimiter=';'):
    """
    Loads a file from the given file_path and loads it into a pandas dataframe
    """
    if not file_path:
        raise ValueError('File path is empty. Please provide a valid path to the file.')

    # Obtain the file extension
    _, file_type = os.path.splitext(file_path)
    # remove the dot from the file_extension
    file_type = file_type[1:]

    # Create a dictionary where the keys are file types and the values are Pandas read functions
    read_funcs = {'csv': lambda x: pd.read_csv(x, delimiter=delimiter),
                  'xls': pd.read_excel,
                  'xlsx': pd.read_excel,
                  'xlsm': pd.read_excel,
                  'json': pd.read_json,
                  'txt': lambda x: pd.read_csv(x, delimiter=delimiter),
                  'h5': pd.read_hdf,
                  'hdf5': pd.read_hdf,
                  'pkl': pd.read_pickle,
                  'pickle': pd.read_pickle,
                  'html': pd.read_html,
                  'htm': pd.read_html}

    try:
        if file_type.lower() not in read_funcs:
            raise ValueError('Unsupported file type. Please try using a file with supported extension.')

        df = read_funcs[file_type.lower()](file_path)

        # If file is HTML, it returns a list of DataFrames. So it will be contacted
        if file_type.lower() in ['html', 'htm']:
            df = pd.concat(df)

        return df

    except FileNotFoundError:
        raise FileNotFoundError(f'File not found at path: {file_path}')

    except Exception as e:
        raise LookupError(f"Error loading the file: {e}")




"""
import logging
import psycopg2
import pymysql
import pyodbc

logging.basicConfig(level=logging.INFO)

def connect_to_database(db_type, host, port, database, user, password):
    conn = None
    try:
        if db_type.lower() == 'postgresql':
            conn = psycopg2.connect(
                host=host,
                port=port,
                database=database,
                user=user,
                password=password
            )
        elif db_type.lower() == 'mysql':
            conn = pymysql.connect(
                host=host,
                port=port,
                database=database,
                user=user,
                password=password
            )
        elif db_type.lower() == 'sqlserver':
            conn = pyodbc.connect(
                'DRIVER={SQL Server};SERVER=' + host + ';DATABASE=' + database + ';UID=' + user + ';PWD=' + password
            )
        else:
            logging.error("Unsupported database type.")
            return None
        
        logging.info(f"Successfully connected to {db_type} database.")
        return conn
    except Exception as e:
        logging.error(f"Error connecting to database: {e}")
        return None
"""




from jetline.pipeline.pipeline import Pipeline
from jetline.pipeline.node import Node
from __PIPE__ import nodes as func

# Dont change name and directory of __file__ (pipline.py)

"""
dont change the name and the directory of the __file__ (pipeline.py) and the structure of the register function.
Just define and order your nodes in the return statement:
Node(name='[can_be_any]', function=func.[your_specific_node], inputs=["Name_of_your_Data_defined_in_data.py"]
, outputs=["Name_of_your_Data_defined_in_data.py"]
"""
def register(data_manager) -> Pipeline:

    return Pipeline(nodes=[
        Node(name='node1', function=func.node_function_1, inputs=["Name"], outputs=["Name"]),
        Node(name='node2', function=func.node_function_2,inputs=["Name"]),
       
    ], data_manager=data_manager)





import os
import json
import importlib.util
import inspect
import ast
import re
from jetline.commands._helper import _extract_pipeline_order,get_project_infos



class FunctionVisitor(ast.NodeVisitor):
    def __init__(self):
        self.functions = []

    def visit_FunctionDef(self, node):
        self.functions.append(node)


def extract_function_code(module_path, function_name):
    spec = importlib.util.spec_from_file_location("module.name", module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    function = getattr(module, function_name)
    return inspect.getsource(function)


def extract_functions_from_module(module_path):
    with open(module_path, "r") as f:
        tree = ast.parse(f.read(), filename=module_path)
        visitor = FunctionVisitor()
        visitor.visit(tree)
        functions = [ast.unparse(func) for func in visitor.functions]
        return functions


def extract_function_code_from_imports(module_path, function_name):
    with open(module_path, "r") as f:
        tree = ast.parse(f.read(), filename=module_path)
        for node in tree.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    try:
                        module = importlib.import_module(alias.name)
                        source_path = inspect.getsourcefile(module)
                        if source_path:
                            functions = extract_functions_from_module(source_path)
                            for func_code in functions:
                                if function_name in func_code:
                                    return func_code
                    except ImportError:
                        pass
            elif isinstance(node, ast.ImportFrom):
                try:
                    module = importlib.import_module(node.module)
                    source_path = inspect.getsourcefile(module)
                    if source_path:
                        functions = extract_functions_from_module(source_path)
                        for func_code in functions:
                            if function_name in func_code:
                                return func_code
                except ImportError:
                    pass
    return None


def extract_register_function(folder_path):
    pipeline_file_path = os.path.join(folder_path, "pipeline.py")

    if os.path.exists(pipeline_file_path):
        with open(pipeline_file_path, "r") as f:
            lines = f.readlines()
            register_function_lines = []
            found_register_function = False
            for line in lines:
                if found_register_function:
                    if line.strip():  # Check if line is not empty
                        register_function_lines.append(line.rstrip())
                    else:
                        break
                elif "def register(" in line:
                    found_register_function = True
                    register_function_lines.append(line.rstrip())
            if register_function_lines:
                register_function = "\n".join(register_function_lines)
                return register_function.strip()

    return None


def extract_node_inputs(node_code):
    inputs = []
    for keyword in node_code.keywords:
        if keyword.arg == "inputs" and isinstance(keyword.value, ast.List):
            for input_elt in keyword.value.elts:
                input_str = ast.unparse(input_elt)
                if isinstance(input_str, str):
                    inputs.append(input_str)
    return inputs


def extract_node_description(function_code):
    if function_code:
        match = re.search(r'"""(.*?)"""', function_code, re.DOTALL)
        if match:
            return match.group(1).strip()
    return None


def extract_register_description(register_function_code):
    if register_function_code:
        match = re.search(r'"""(.*?)"""', register_function_code, re.DOTALL)
        if match:
            return match.group(1).strip()
    return None



def extract_functions_from_register(register_function):
    """Extract functions, inputs, outputs from a given register function."""

    functions = re.findall(r'function=([a-zA-Z0-9_.-]+)', register_function)
    inputs = re.findall(r'inputs=\[(.*?)\]', register_function)
    outputs = re.findall(r'outputs=\[(.*?)\]', register_function)
  
    return functions, inputs, outputs


def format_node_parameters(param):
    """Format given node parameter by removing unnecessary characters and splitting by comma."""

    formatted = param.replace('"', '').replace("'", "").replace(" ", "")
    return formatted.split(',')
def extract_nodes_info(folder_path, register_function):
    """Extract and return nodes information from a given Python file in a folder using a register function."""

    function_names, inputs, outputs = extract_functions_from_register(register_function)

    sorted_nodes = [None] * len(function_names)
    nodes_dict = {}

    node_file_path = os.path.join(folder_path, "nodes.py")
    if os.path.exists(node_file_path):
        with open(node_file_path, "r") as file:
            tree = ast.parse(file.read(), filename=node_file_path)

            for node_def in tree.body:
                if isinstance(node_def, ast.FunctionDef):
                    node_name = node_def.name
                    node_description = ast.get_docstring(node_def) or ''
                    node_code = ast.unparse(node_def)

                    nodes_dict[node_name] = {
                        "data": {
                          
                        "title": node_name,
                          "type": "function",
                        "code": node_code,
                        "description": node_description,
                        "inputs": "",
                        "outputs": "",
                        "stream": ""
                        }
                    }
                for i, func_name in enumerate(function_names):
                    node_name = func_name.split('.')[-1]
                    if node_name in nodes_dict:
                        sorted_nodes[i] = nodes_dict[node_name]

                        if i < len(inputs):
                            sorted_nodes[i]["data"]["inputs"] = format_node_parameters(inputs[i])

                        if i < len(outputs):
                            sorted_nodes[i]["data"]["outputs"] = format_node_parameters(outputs[i])

                        # Adding 'stream' attribute
                        if sorted_nodes[i]["data"]["inputs"] and sorted_nodes[i]["data"]["outputs"]:
                            sorted_nodes[i]["data"]["stream"] = ["input", "output"]
                        elif sorted_nodes[i]["data"]["inputs"]:
                            sorted_nodes[i]["data"]["stream"] = ["input"]
                        elif sorted_nodes[i]["data"]["outputs"]:
                            sorted_nodes[i]["data"]["stream"] = ["output"]
                        else:
                            sorted_nodes[i]["data"]["stream"] = []


            return sorted_nodes

    return []




def extract_register_info(folder_path):
    pipeline_file_path = os.path.join(folder_path, "pipeline.py")

    if os.path.exists(pipeline_file_path):
        with open(pipeline_file_path, "r") as f:
            tree = ast.parse(f.read(), filename=pipeline_file_path)
            for node_def in tree.body:
                if isinstance(node_def, ast.FunctionDef) and node_def.name == "register":
                    register_function_code = ast.unparse(node_def)
                    register_description = extract_register_description(register_function_code)
                    return register_description

    return None

def extract_data_classes_info(current_directory):
    data_classes_info = []
    data_file_path = os.path.join(current_directory, "data.py")
    if os.path.exists(data_file_path):
        with open(data_file_path, 'r') as f:
            tree = ast.parse(f.read(), filename=data_file_path)
            for i,node in enumerate(tree.body):
                if isinstance(node, ast.ClassDef):
                    class_description = ast.get_docstring(node)
                    class_code = ast.unparse(node)

                    class_name_match = re.search(r"name\s*=\s*'([^']*)'", class_code)
                    class_name = class_name_match.group(1) if class_name_match else "Unknown"

                    data_classes_info.append({
                      
                        "data": {
                        "title": class_name,
                         "type": "data",
                        "description": class_description if class_description else "",
                        "code": class_code}
                    })
    return data_classes_info




def main():
    
    project_name,place_path, current_dir=  get_project_infos()
    

   
    pipeline_order = _extract_pipeline_order(current_dir)
    pipeline_folders = [folder for folder in os.listdir(place_path) if os.path.isdir(os.path.join(place_path, folder))]
    pipeline_folders = sorted(pipeline_folders,
                              key=lambda x: pipeline_order.index(x) if x in pipeline_order else len(pipeline_order))
    
    nodes = []
    data = {
        "ProjectName": project_name,
        "DataClasses": [],
        "Pipelines": {}
    }

    # Extract data classes information
    
    nodes.extend(extract_data_classes_info(current_dir))

    
   # data["DataClasses"] = data_classes_info

    # Extract pipeline information

    for folder in pipeline_folders:
        folder_path = os.path.join(place_path, folder)
        pipeline_info = {
            "RegisterFunction": {
                "Function": None,
                "Description": None
            },
            "Nodes": None
        }

        # Extract register function information
        register_function = extract_register_function(folder_path)
        pipeline_info["RegisterFunction"]["Function"] = extract_register_function(folder_path)
        pipeline_info["RegisterFunction"]["Description"] =  extract_register_description(register_function)

        # Extract nodes information
        nodes_info = extract_nodes_info(folder_path, register_function)
        pipeline_info["Nodes"] = nodes_info
        nodes.extend(nodes_info)
      
        # Add pipeline information to data
        data["Pipelines"][folder] = pipeline_info
    
  #  print(nodes)
    json_file_path = os.path.join(current_dir, "viz-data.json")
    with open(json_file_path, "w") as json_file:
        json.dump(nodes, json_file, indent=4)

    print("JSON-Datei erfolgreich erstellt.")


if __name__ == "__main__":
    main()



import re
import os
import click
import toml
from colorama import Fore, Style
def echo_jetline():

    click.echo(f"\n")
    click.echo(f"{Fore.BLUE}░▀▀█░█▀▀░▀█▀░█░░░▀█▀░█▀█░█▀▀{Style.RESET_ALL}")
    click.echo(f"{Fore.BLUE}░░░█░█▀▀░░█░░█░░░░█░░█░█░█▀▀{Style.RESET_ALL}")
    click.echo(f"{Fore.BLUE}░▀▀░░▀▀▀░░▀░░▀▀▀░▀▀▀░▀░▀░▀▀▀{Style.RESET_ALL}")
    click.echo(f"\n... a powerful, lightweight pipeline builder! by {Fore.GREEN}Kdc-Solutions{Style.RESET_ALL}\n")
    
    
def get_project_infos():
    """Fetches the project name and path based on project settings in a TOML file."""
    current_dir = os.getcwd()
    toml_file_path = os.path.join(current_dir, "project.toml")

    try:
        with open(toml_file_path, "r") as file:
            toml_data = toml.load(file)
            project_name = toml_data.get("project", {}).get("name")
            place = toml_data.get("project", {}).get("place")
            if not project_name or not place:
                return None, None
            place_path = os.path.join(current_dir, place)
            return project_name, place_path, current_dir
    except Exception:
        raise ValueError("Project name or place is missing in the TOML file.")


def _extract_pipeline_order(current_directory):
    """
    Extracts the PIPELINE_ORDER list from the content of the main file.

    :param current_directory: The directory of the main file.
    :return: The extracted PIPELINE_ORDER list or None if not found.
    """
    main_path = os.path.join(current_directory, 'main.py')
    with open(main_path, 'r') as file:
        main_content = file.read()

    pipeline_order_match = re.search(r'PIPELINE_ORDER\s*=\s*\[([^\]]+)\]', main_content)
    if pipeline_order_match:
        return pipeline_order_match.group(1).replace("'", "").replace('"', "").split(',')
    return None



def node_function_1(data):
    """
    An example function that processes data and returns a modified version.

    Args:
        data: The input data to be processed.

    Returns:
        str: The processed data.
    """
    print("Input Data:", data)
    processed_data = data + "asasdasdd"
    print("Processed Data:", processed_data)
    return processed_data

def node_function_2(data):
    """
    An example function that returns the input data unchanged.

    Args:
        data: The input data.

    Returns:
        str: The input data.
    """
    print("Input Data:", data)
    return data
